@@ -1,6 +1,6 @@
 /******************************************************************************\
- * $Id: buffer.cpp,v 1.4 2001/05/28 14:39:15 blais Exp $
- * $Date: 2001/05/28 14:39:15 $
+ * $Id: buffer.cpp,v 1.33 2002/02/23 17:43:56 blais Exp $
+ * $Date: 2002/02/23 17:43:56 $
  *
  * Copyright (C) 1999-2001  Martin Blais <blais@iro.umontreal.ca>
  *
@@ -18,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *****************************************************************************/
+ ******************************************************************************/
 
 /*==============================================================================
  * EXTERNAL DECLARATIONS
@@ -30,32 +30,17 @@
 #include <qfontmetrics.h>
 #include <qfont.h>
 #include <qrect.h>
+#include <qdir.h>
 
 #include <iostream>
-#include <string>
+#include <string.h>
 #include <algorithm>
+
 #include <unistd.h>
 #include <stdio.h>
 #include <math.h>
 #include <errno.h>
 
-#if HAVE_DIRENT_H
-# include <dirent.h>
-# define NAMLEN(dirent) strlen((dirent)->d_name)
-#else    
-# define dirent direct
-# define NAMLEN(dirent) ((dirent)->d_namlen)
-# if HAVE_SYS_NDIR_H
-#  include <sys/ndir.h>
-# endif
-# if HAVE_SYS_DIR_H
-#  include <sys/dir.h>
-# endif
-# if HAVE_NDIR_H
-#  include <ndir.h>
-# endif
-#endif
-
 /*==============================================================================
  * LOCAL DECLARATIONS
  *============================================================================*/
@@ -76,21 +61,29 @@ const char* strnchr( const char* s, int c, const char* end )
    return 0;
 }
 
-
 //------------------------------------------------------------------------------
 //
-bool compareStrings( const char* s1, const char* s2 ) 
+const char* strnstr( const char* haystack, const uint n, const char* needle ) 
 {
-   XX_CHECK( s1 != 0 && s2 != 0 );
-   return ::strcmp( s1, s2 ) < 0;
+   const char* p = haystack;
+   const char* pe = p + n - ::strlen( needle );
+   while ( p < pe ) {
+      while ( *p != *needle && p < pe ) { ++p; }
+      const char* pp = p;
+      const char* nn = needle;
+      while ( (*pp == *nn) && (*nn != 0) ) { ++pp; ++nn; }
+      if ( *nn == 0 ) {
+         return p;
+      }
+      ++p;
+   }
+   return 0;
 }
 
 }
 
-
 XX_NAMESPACE_BEGIN
 
-
 /*==============================================================================
  * PUBLIC FUNCTIONS
  *============================================================================*/
@@ -102,14 +95,15 @@ XX_NAMESPACE_BEGIN
 //------------------------------------------------------------------------------
 //
 XxBuffer::XxBuffer( 
-   const bool  passiveDummy, // ignored
-   const char* filename, 
-   const char* displayFilename
+   const bool     /*passiveDummy*/, // ignored
+   const QString& filename,
+   const QString& displayFilename
 ) :
    _name( filename ),
    _displayName( displayFilename ),
    _hiddenCR( false ),
    _temporary( false ),
+   _proxy( false ),
    _buffer( 0 )
 {
    XX_ASSERT( filename != 0 );
@@ -120,35 +114,60 @@ XxBuffer::XxBuffer(
 //------------------------------------------------------------------------------
 //
 XxBuffer::XxBuffer( 
-   const char* filename, 
-   const char* displayFilename,
-   const bool  hideCR,
-   const bool  deleteFile
+   const QString& filename, 
+   const QString& displayFilename,
+   const bool     hideCR,
+   const bool     deleteFile
 ) :
    _name( filename ),
    _displayName( displayFilename ),
    _hiddenCR( hideCR ),
    _temporary( deleteFile ),
+   _proxy( false ),
    _buffer( 0 )
 {
-   XX_ASSERT( filename != 0 );
+   XX_ASSERT( !filename.isEmpty() );
 
    // Stat the file.
-   struct stat ss;
-   if ( stat( _name.c_str(), &ss ) == -1 ) {
-      throw new XxIoError;
+   QFileInfo finfo( _name );
+   if ( !finfo.exists() ) {
+      throw XxIoError( XX_EXC_PARAMS );
    }
 
-   if ( ! S_ISDIR( ss.st_mode ) ) {
-      loadFile( ss );
+   if ( finfo.isDir() ) {
+      loadDirectory();
    }
    else {
-      loadDirectory( ss );
+      loadFile( finfo );
    }
 
    init();
 }
 
+//------------------------------------------------------------------------------
+//
+XxBuffer::XxBuffer(
+   const XxBuffer& orig,
+   const QString&  filename,
+   const QString&  displayFilename
+) :
+   _name( filename ),
+   _displayName( displayFilename ),
+   _hiddenCR( orig._hiddenCR ),
+   _temporary( false ),
+   _proxy( true ),
+   _buffer( orig._buffer ), // Here is the "sharing the buffer" part.
+   _bufferSize( orig._bufferSize ),
+   _index( orig._index )
+#ifdef XX_ENABLED_BUFFER_LINE_LENGTHS
+   , _lengths( orig._lengths )
+#endif
+{
+   XX_ASSERT( filename != 0 );
+
+   init();
+}
+
 //------------------------------------------------------------------------------
 //
 void XxBuffer::init()
@@ -161,48 +180,51 @@ void XxBuffer::init()
 //
 XxBuffer::~XxBuffer()
 {
-   delete _buffer;
+   if ( !_proxy ) {
+      delete[] _buffer;
+   }
+   _buffer = 0;
 
    // Delete the temporary file if asked for.
    if ( _temporary == true ) {
-      unlink( _name.c_str() );
+      unlink( _name.latin1() );
    }
 
    if ( _renderBuffer != 0 ) {
       free( _renderBuffer );
    }
-
-   // Cleanup string of directory entries.
-   for ( uint ii = 0; ii < _directoryEntries.size(); ++ii ) {
-      if ( _directoryEntries[ii] != 0 ) {
-         delete[] _directoryEntries[ii];
-      }
-   }
 }
 
 //------------------------------------------------------------------------------
 //
-void XxBuffer::loadFile( struct stat& ss )
+void XxBuffer::setDisplayName( const QString& fn )
+{
+   _displayName = fn;
+}
+
+//------------------------------------------------------------------------------
+//
+void XxBuffer::loadFile( const QFileInfo& finfo )
 {
    // Find out the file size.
-   _bufferSize = ss.st_size;
+   _bufferSize = finfo.size();
 
    // Allocate buffer.
    // add one for potential added newline.
    _buffer = new char[ _bufferSize + 1 ];
 
    // Read file into buffer.
-   FILE* fp = fopen( _name.c_str(), "r" );
+   FILE* fp = fopen( _name.latin1(), "r" );
    if ( fp == 0 ) {
-      throw new XxIoError;
+      throw XxIoError( XX_EXC_PARAMS );
    }
 
    if ( fread( _buffer, 1, _bufferSize, fp ) != _bufferSize ) {
-      throw new XxIoError;
+      throw XxIoError( XX_EXC_PARAMS );
    }
 
    if ( fclose( fp ) != 0 ) {
-      throw new XxIoError;
+      throw XxIoError( XX_EXC_PARAMS );
    }
 
    // Add a final newline if there isn't one.  This will simplify indexing.
@@ -222,72 +244,42 @@ void XxBuffer::loadFile( struct stat& ss )
 
 //------------------------------------------------------------------------------
 //
-void XxBuffer::loadDirectory( struct stat& ss )
+void XxBuffer::loadDirectory()
 {
-   register struct dirent* next;
-   using namespace std;
-
-   std::vector<const char*> directoryEntries;
-
-   /* Open the directory and check for errors.  */
-   register DIR* reading = opendir( _name.c_str() );
-   if ( !reading ) {
-      throw new XxIoError;
-   }
-   
-   /* Read the directory entries, and insert the subfiles
-      into the `data' table.  */
-   while ( ( errno = 0, ( next = readdir( reading ) ) != 0 ) ) {
-      char *d_name = next->d_name;
-      size_t d_size = NAMLEN(next) + 1;
-         
-      /* Ignore the files `.' and `..' */
-      if ( d_name[0] == '.'
-           && (d_name[1] == 0 || (d_name[1] == '.' && d_name[2] == 0))) {
-         continue;
-      }
-         
-      char* newString = new char[ d_size + 1 ];
-      ::strncpy( newString, d_name, d_size );
-      newString[ d_size ] = 0;
-
-      XX_CHECK( newString != 0 && d_name != 0 && d_size > 0 );
-      directoryEntries.push_back( newString );
-   }
-   if ( errno ) {
-      int e = errno;
-      closedir( reading );
-      errno = e;
-      throw new XxIoError;
-   }
-#if CLOSEDIR_VOID
-   closedir( reading );
-#else
-   if ( closedir( reading ) != 0 ) {
-      throw new XxIoError;
-   }
-#endif
-
-   std::sort(
-      directoryEntries.begin(),
-      directoryEntries.end(),
-      compareStrings
-   );
+   QDir dir( _name );
+   dir.setFilter( QDir::Dirs | QDir::Files | QDir::Hidden | QDir::System );
+// Dirs - List directories only 
+// Files - List files only 
+// Drives - List disk drives (does nothing under unix) 
+// NoSymLinks - Do not list symbolic links (where they exist) 
+// Readable - List files for which the application has read access. 
+// Writable - List files for which the application has write access. 
+// Executable - List files for which the application has execute access 
+// Modified - Only list files that have been modified (does nothing under unix) 
+// Hidden - List hidden files (on unix, files starting with a .) 
+// System - List system files (does nothing under unix) 
 
+   dir.setSorting( QDir::Unsorted );
+   XX_CHECK( dir.exists() );
+   QStringList directoryEntries = dir.entryList();
+   directoryEntries.remove( "." );
+   directoryEntries.remove( ".." );
    setDirectoryEntries( directoryEntries );
 }
 
 //------------------------------------------------------------------------------
 //
 void XxBuffer::setDirectoryEntries( 
-   const std::vector<const char*>& directoryEntries 
+   const QStringList& directoryEntries 
 )
 {
    _directoryEntries = directoryEntries;
 
    _bufferSize = 0;
-   for ( uint ii = 0; ii < _directoryEntries.size(); ++ii ) {
-      _bufferSize += ::strlen( _directoryEntries[ii] ) + 1;
+   for ( QStringList::Iterator itr = _directoryEntries.begin(); 
+         itr != _directoryEntries.end();
+         ++itr ) {
+      _bufferSize += (*itr).length() + 1;
    }
 
    // Allocate buffer.
@@ -295,9 +287,11 @@ void XxBuffer::setDirectoryEntries(
    _buffer = new char[ _bufferSize + 1 ];
    
    char* bufferPtr = _buffer;
-   for ( uint ii = 0; ii < _directoryEntries.size(); ++ii ) {
-      int len = ::strlen( _directoryEntries[ii] );
-      ::strncpy( bufferPtr, _directoryEntries[ii], len );
+   for ( QStringList::Iterator it = _directoryEntries.begin(); 
+         it != _directoryEntries.end();
+         ++it ) {
+      int len = (*it).length();
+      ::strncpy( bufferPtr, (*it).latin1(), len );
       bufferPtr[len] = '\n';
       bufferPtr += len + 1;
    }
@@ -340,27 +334,76 @@ void XxBuffer::processCarriageReturns()
 void XxBuffer::indexFile()
 {
    XX_ASSERT( _buffer );
+
    uint ii;
+   _index.push_back( -1 );
    _index.push_back( 0 );
+#ifdef XX_ENABLED_BUFFER_LINE_LENGTHS
+   int prev = 0;
+   _lengths.push_back( -1 );
+#endif
 
    for ( ii = 0; ii < _bufferSize; ++ii ) {
       if ( _buffer[ii] == '\n' ) {
          _index.push_back( ii + 1 );
+#ifdef XX_ENABLED_BUFFER_LINE_LENGTHS
+         _lengths.push_back( ii - prev );
+         prev = static_cast<short>( ii + 1 );
+#endif
       }
    }
+   _lengths.push_back( 0 ); // Just so that the two arrays are the same size.
+#ifdef XX_ENABLED_BUFFER_LINE_LENGTHS
+   XX_ASSERT( _index.size() == _lengths.size() );
+#endif
+
    // Make sure the file was [possibly artificially] terminated with a newline.
    XX_ASSERT( _index.back() == int(_bufferSize) );
 }
 
+//------------------------------------------------------------------------------
+//
+void XxBuffer::reindex( const std::vector<XxFln>& reindexTbl )
+{
+   XX_ASSERT( _buffer );
+
+   // Swap old indexes.
+   std::vector<int> oldIndex;
+   std::swap( _index, oldIndex );
+
+#ifdef XX_ENABLED_BUFFER_LINE_LENGTHS
+   std::vector<short> oldLengths;
+   std::swap( _lengths, oldLengths );
+#endif
+
+   uint ii;
+   _index.push_back( -1 );
+#ifdef XX_ENABLED_BUFFER_LINE_LENGTHS
+   _lengths.push_back( -1 );
+#endif
+   
+   for ( ii = 1; ii < reindexTbl.size(); ++ii ) {
+      int reidx = reindexTbl[ii];
+      _index.push_back( oldIndex[reidx] );
+#ifdef XX_ENABLED_BUFFER_LINE_LENGTHS
+      _lengths.push_back( oldLengths[reidx] );
+#endif
+   }
+
+   _dpyLineNos = reindexTbl;
+}
+
 //------------------------------------------------------------------------------
 //
 std::ostream& XxBuffer::dump( std::ostream& os ) const
 {
+   os << "Buffer dump:" << std::endl;
+   os << "------------------------------" << std::endl;
    int nbLines = getNbLines();
    for ( XxFln ii = 1; ii <= nbLines; ++ii ) {
       uint length;
       const char* line = getTextLine( ii, length );
-      os << ii << ":";
+      os << ii << "(" << length << ")[" << ( line - _buffer ) << "]:";
       os.write( line, length );
       os << std::endl;
    }
@@ -380,14 +423,14 @@ uint XxBuffer::computeTextWidth( const QFont& font, const uint tabWidth )
       const char* lineText = getTextLine( ii, length );
 
       int rlength;
-      int lhd = 0;
-      int rhd = 0;
       const char* renderedText = 
-         renderTextWithTabs( lineText, length, tabWidth, rlength, lhd, rhd );
+         renderTextWithTabs( lineText, length, tabWidth, rlength, 0 );
       QString str( renderedText );
 
       QRect rect = fm.boundingRect( str );
-      longest = std::max( longest, (uint)rect.width() );
+      if ( static_cast<uint>( rect.width() ) > longest ) {
+         longest = static_cast<uint>( rect.width() );
+      }
    }
    return longest;
 }
@@ -404,7 +447,8 @@ uint XxBuffer::computeLineNumbersWidth( const QFont& font ) const
    int nbLines = getNbLines();
    char buffer[12];
    for ( int ii = 1; ii <= nbLines; ++ii ) {
-      ::snprintf( buffer, sizeof(buffer), lnFormat, ii );
+      XxFln fline = getDisplayLineNo( ii );
+      ::snprintf( buffer, sizeof(buffer), lnFormat, fline );
       QString str( buffer );
 
       QRect rect = fm.boundingRect( str, str.length() );
@@ -420,8 +464,7 @@ const char* XxBuffer::renderTextWithTabs(
    const uint  length,
    const uint  tabWidth,
    int&        rlength,
-   int&        lhd,
-   int&        rhd
+   int*        hordiffs
 )
 {
    const int increment = 256;
@@ -434,7 +477,6 @@ const char* XxBuffer::renderTextWithTabs(
    const char* nexttab = 0;
 
    // Offset rightmost char to render tabs correctly.
-   int hrhd = rhd + 1;
    while ( ( nexttab = strnchr( ps, '\t', endstr ) ) != 0 ) {
 
       // Check for size requirements and reallocate if necessary.
@@ -452,21 +494,31 @@ const char* XxBuffer::renderTextWithTabs(
       }
       ps++; // skip tab
 
-      // Output tab.
+      // Compute amount of spaces to insert.
       uint nspaces = 0;
       if ( tabWidth != 0 ) {
          nspaces = tabWidth - ((pd - _renderBuffer) % tabWidth);
       }
 
-      lhd += ( lhd > (pd - _renderBuffer) ) ? nspaces - 1 : 0;
-      hrhd += ( hrhd > (pd - _renderBuffer) ) ? nspaces - 1 : 0;
+      // Shift horizontal diffs if necessary.
+      if ( hordiffs ) {
 
+         for ( int* phd = hordiffs; *phd != -1; ++phd ) {
+            if ( *phd > (pd - _renderBuffer) ) {
+               *phd += nspaces - 1;
+               // -1 because we the hordiffs were calculated with a string
+               // -containing tab characters in them, and the tab chars take 1
+               // -char space in the string.
+            }
+         }
+      }
+
+      // Output tab.
       for ( uint t = 0; t < nspaces; ++t ) {
          *pd++ = ' ';
          XX_CHECK( pd - _renderBuffer < _renderBufferSize );
       }
    }
-   rhd = hrhd - 1;
 
    // Copy last chunk.
 
@@ -499,28 +551,25 @@ uint XxBuffer::getNbDigits() const
 
 //------------------------------------------------------------------------------
 //
-const char* XxBuffer::renderLineNumber( 
-   const XxFln lineNumber,
-   const char* format
+const QString& XxBuffer::renderLineNumber( 
+   const XxFln    lineNumber,
+   const QString& format
 )
 {
-   ::snprintf( _lnBuffer, sizeof(_lnBuffer), format, lineNumber );
+   _lnBuffer.sprintf( format, lineNumber );
    return _lnBuffer;
 }
 
 //------------------------------------------------------------------------------
 //
-bool XxBuffer::searchLine( const XxFln lineno, const char* searchText ) const
+bool XxBuffer::searchLine( const XxFln lineno, const QString& searchText ) const
 {
    bool found = false;
    uint len;
-   char* text = const_cast<char*>( getTextLine( lineno, len ) );
-   char endchar = text[len];
-   text[len] = '\0';
-   if ( strstr( text, searchText ) != 0 ) {
+   const char* text = getTextLine( lineno, len );
+   if ( strnstr( text, len, searchText.latin1() ) != 0 ) {
       found = true;
    }
-   text[len] = endchar;
    return found;
 }
 
@@ -528,28 +577,30 @@ bool XxBuffer::searchLine( const XxFln lineno, const char* searchText ) const
 //
 bool XxBuffer::isDirectory() const
 {
-   return _directoryEntries.size() > 0;
+   return !_directoryEntries.isEmpty();
 }
 
 //------------------------------------------------------------------------------
 //
-const std::vector<const char*>& XxBuffer::getDirectoryEntries() const
+const QStringList& XxBuffer::getDirectoryEntries() const
 {
    return _directoryEntries;
 }
 
 //------------------------------------------------------------------------------
 //
-std::string XxBuffer::getFileAtLine( 
+QString XxBuffer::getBufferAtLine( 
    const XxFln lineno
 ) const
 {
-   std::string filename( _name );
-   if ( filename[ filename.length()-1 ] != '/' ) {
+   QString filename = _name;
+   if ( filename.constref( filename.length()-1 ) != '/' ) {
       filename.append( "/" );
    }
    
-   XX_ASSERT( lineno - 1 < XxFln(_directoryEntries.size()) );
+   XX_CHECK( lineno - 1 < XxFln(_directoryEntries.count()) );
+
+   // Note: QStringList O(n) lookup.
    filename.append( _directoryEntries[ lineno - 1 ] );
    return filename;
 }
