@@ -1,6 +1,6 @@
 /******************************************************************************\
- * $Id: builderDirs2.cpp,v 1.5 2001/05/28 18:32:02 blais Exp $
- * $Date: 2001/05/28 18:32:02 $
+ * $Id: builderDirs2.cpp,v 1.27 2002/02/23 17:43:56 blais Exp $
+ * $Date: 2002/02/23 17:43:56 $
  *
  * Copyright (C) 1999-2001  Martin Blais <blais@iro.umontreal.ca>
  *
@@ -18,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *****************************************************************************/
+ ******************************************************************************/
 
 /*==============================================================================
  * EXTERNAL DECLARATIONS
@@ -30,14 +30,25 @@
 #include <util.h>
 #include <buffer.h>
 
+#include <qstring.h>
+#include <qstringlist.h>
+#include <qtextstream.h>
+#include <qfile.h>
+
+#include <stdexcept>
 #include <stdio.h>
 #include <stdlib.h>
-#include <sstream>
+#include <string.h>
 #include <algorithm>
+
 #include <unistd.h>
+#ifndef WINDOWS
+#include <sys/wait.h>
+#else
+#endif
+
 #include <iostream>
 #include <sys/types.h>
-#include <sys/wait.h>
 
 //#define LOCAL_TRACE
 #ifdef LOCAL_TRACE
@@ -71,29 +82,12 @@ char* typeString[5] = {
 
 /*----- classes -----*/
 
-/*==============================================================================
- * STRUCT StringEqual
- *============================================================================*/
-
-struct StringEqual {
-
-   StringEqual( const char* s ) :
-      _s( s )
-   {}
-
-   bool operator () ( const char* s ) {
-      return ::strcmp( _s, s ) == 0;
-   }
-
-   const char* _s;
-
-};
-
 /*==============================================================================
  * CLASS XxParseDiffError
  *============================================================================*/
 
-class XxParseDiffError : public std::exception {
+class XxParseDiffError : public XxError,
+                         public std::runtime_error {
 
 public:
 
@@ -101,23 +95,10 @@ public:
 
    // Constructor with state.
    XxParseDiffError(
-      const char* buf,
-      const char* file, 
-      int         lineno
+      XX_EXC_PARAMS_DECL(file,line),
+      const char* buf
    );
 
-   // Destructor.
-   virtual ~XxParseDiffError() XX_THROW_NOTHING;
-
-   // See base class.
-   virtual const char* what() const XX_THROW_NOTHING;
-
-private:
-
-   /*----- data members -----*/
-
-   std::string _msg;
-
 };
 
 /*==============================================================================
@@ -127,29 +108,15 @@ private:
 //------------------------------------------------------------------------------
 //
 XxParseDiffError::XxParseDiffError(
-   const char* buf,
-   const char* file, 
-   int         lineno
-)
+   XX_EXC_PARAMS_DECL(file,line),
+   const char* buf
+) :
+   XxError( file, line ),
+   std::runtime_error( "Parse diff output error." )
 {
-   std::ostringstream oss;
-   oss << "Error parsing diff output: " << std::endl
-       << buf << std::endl
-       << "File: " << file << "Line: " << lineno << std::endl
-       << std::ends;
-   _msg = oss.str();
-}
-
-//------------------------------------------------------------------------------
-//
-XxParseDiffError::~XxParseDiffError() XX_THROW_NOTHING
-{}
-
-//------------------------------------------------------------------------------
-//
-const char* XxParseDiffError::what() const XX_THROW_NOTHING
-{
-   return _msg.c_str();
+   QTextStream oss( &_msg, IO_WriteOnly | IO_Append );
+   oss << "Error parsing diff output: " << endl
+       << buf << endl;
 }
 
 /*==============================================================================
@@ -159,28 +126,31 @@ const char* XxParseDiffError::what() const XX_THROW_NOTHING
 //------------------------------------------------------------------------------
 //
 bool parseDiffLine( 
-   const char*  buf,
-   const char*  dir1,
-   int          len1,
-   const char*  dir2,
-   int          len2,
-   DirDiffType& type,
-   std::string& filename,
-   int&         onlyDir
+   const QString& line,
+   const QString& dir1,
+   int            len1,
+   const QString& dir2,
+   int            len2,
+   DirDiffType&   type,
+   QString&       filename,
+   int&           onlyDir
 )
 {
+   const char* buf = line.latin1();
+
    bool error;
    const char* bufPtr = buf;
+   onlyDir = -1; /* don't take chances with chance */
    if ( ::strncmp( bufPtr, "Only in ", 8 ) == 0 ) {
       bufPtr += 8;
       const char* colonPtr = ::strstr( bufPtr, ": " );
       if ( colonPtr == 0 ) {
-         throw new XxParseDiffError( buf, XX_INTERROR_PARAMS );
+         throw XxParseDiffError( XX_EXC_PARAMS, buf );
       }
-
+      
       //int len = colonPtr - bufPtr;
-      int cmp1 = ::strncmp( bufPtr, dir1, len1 );
-      int cmp2 = ::strncmp( bufPtr, dir2, len2 );
+      int cmp1 = ::strncmp( bufPtr, dir1.latin1(), len1 );
+      int cmp2 = ::strncmp( bufPtr, dir2.latin1(), len2 );
       // Note: you cannot compare the lengths because these might be directory
       // diffs.
       if ( cmp1 == 0 && cmp2 == 0 ) {
@@ -188,32 +158,37 @@ bool parseDiffLine(
          // the directories is a subdirectory of the other, but the output is
          // ambiguous, and to really check we would have to stat. Besides GNU
          // diff leaves this ambibuity.
-         onlyDir = cmp1 > cmp2 ? 0 : 1;
+         onlyDir = (len1 > len2) ? 0 : 1;
       }
       else if ( cmp1 != 0 && cmp2 != 0 ) {
-         throw new XxParseDiffError( buf, XX_INTERROR_PARAMS );
+         throw XxParseDiffError( XX_EXC_PARAMS, buf );
       }
       else {
-         onlyDir = cmp1 == 0 ? 0 : 1;
+         onlyDir = (cmp1 == 0) ? 0 : 1;
       }
 
       int lens = ( onlyDir == 0 ? len1 : len2 );
       // Note: skip the "/" after the specified directory path
-      filename = std::string(); // Note: clear() ins't implemented in libstdc++.
+      filename = QString();
 
       const char* dnamePtr = bufPtr + lens;
       if ( *dnamePtr == '/' ) {
          ++dnamePtr;
       }
       if ( colonPtr - dnamePtr > 0 ) {
-         filename += std::string( dnamePtr, colonPtr - dnamePtr );
-         if ( filename[ filename.length() - 1 ] != '/' ) {
-            filename += "/";
+         QString aname; 
+         aname.setLatin1( dnamePtr, colonPtr - dnamePtr );
+         filename += aname;
+         if ( filename.constref( filename.length() - 1 ) != '/' ) {
+            filename.append( '/' );
          }
       }
 
       // Note: need to remove \n
-      filename += std::string( colonPtr + 2, ::strlen( colonPtr + 2 ) - 1 );
+      QString bname;
+      bname.setLatin1( colonPtr + 2 );
+
+      filename += bname;
       type = ONLY_IN;
       error = false;
    }
@@ -221,13 +196,19 @@ bool parseDiffLine(
       bufPtr += 6;
       const char* andPtr = ::strstr( bufPtr, " and " );
       if ( andPtr == 0 ) {
-         throw new XxParseDiffError( buf, XX_INTERROR_PARAMS );
+         throw XxParseDiffError( XX_EXC_PARAMS, buf );
       }
       const char* filenamePtr = bufPtr + len1;
       if ( *filenamePtr == '/' ) {
          ++filenamePtr;
       }
-      filename = std::string( filenamePtr, andPtr - filenamePtr );
+      int mlen = andPtr - filenamePtr;
+      if ( mlen > 0 ) {
+         filename.setLatin1( filenamePtr, mlen );
+      }
+      else {
+         filename = QString();
+      }
 
       bufPtr = andPtr + 5;
       const char* endtagPtr = ::strstr( bufPtr, " differ" );
@@ -238,7 +219,7 @@ bool parseDiffLine(
          type = IDENTICAL;
       }
       else {
-         throw new XxParseDiffError( buf, XX_INTERROR_PARAMS );
+         throw XxParseDiffError( XX_EXC_PARAMS, buf );
       }
       error = false;
    }
@@ -246,13 +227,19 @@ bool parseDiffLine(
       bufPtr += 23;
       const char* andPtr = ::strstr( bufPtr, " and " );
       if ( andPtr == 0 ) {
-         throw new XxParseDiffError( buf, XX_INTERROR_PARAMS );
+         throw XxParseDiffError( XX_EXC_PARAMS, buf );
       }
       const char* filenamePtr = bufPtr + len1;
       if ( *filenamePtr == '/' ) {
          ++filenamePtr;
       }
-      filename = std::string( filenamePtr, andPtr - filenamePtr );
+      int mlen = andPtr - filenamePtr;
+      if ( mlen > 0 ) {
+         filename.setLatin1( filenamePtr, mlen );
+      }
+      else {
+         filename = QString();
+      }
 
       type = COMMON_SUBDIR;
       error = false;
@@ -264,50 +251,28 @@ bool parseDiffLine(
    return error;
 }
 
-//------------------------------------------------------------------------------
-//
-int searchEntry( 
-   const std::vector<const char*>& entries, 
-   const char*                     filename
-)
-{
-   std::vector<const char*>::const_iterator iter = 
-      std::find_if( 
-         entries.begin(), 
-         entries.end(),
-         StringEqual( filename )
-      );
-
-   if ( iter == entries.end() ) {
-      return -1;
-   }
-
-   return iter - &entries[0];
-}
-
 //------------------------------------------------------------------------------
 //
 void setType( 
-   const std::vector<const char*>& entries, 
-   std::vector<DirDiffType>&       types, 
-   const char*                     filename, 
-   DirDiffType                     type
+   const QStringList&        entries, 
+   std::vector<DirDiffType>& types, 
+   const QString&            filename, 
+   DirDiffType               type
 )
 {
-   int index = searchEntry( entries, filename );
-
+   int index = entries.findIndex( filename );
    if ( index == -1 ) {
 #ifdef LOCAL_TRACE
-      for ( int ii = 0; ii < entries.size(); ++ii ) {
-         std::cout << entries[ii] << std::endl;
+      for ( unsigned int ii = 0; ii < entries.count(); ++ii ) {
+         XX_TRACE( entries[ii] );
       }
-      std::cout << "filename \"" << filename << "\"" << std::endl;
+      XX_TRACE( "filename \"" << filename.latin1() << "\"" );
 #endif
-      throw new XxInternalError( XX_INTERROR_PARAMS );
+      throw XxInternalError( XX_EXC_PARAMS );
    }
 
    if ( index >= int( types.size() ) ) {
-      throw new XxInternalError( XX_INTERROR_PARAMS );
+      throw XxInternalError( XX_EXC_PARAMS );
    }
    types[ index ] = type;
 }
@@ -315,15 +280,15 @@ void setType(
 //------------------------------------------------------------------------------
 //
 void patchUpMissingTypes( 
-   const std::vector<const char*>& entries1,
-   const std::vector<const char*>& entries2,
-   std::vector<DirDiffType>&       types1,
-   std::vector<DirDiffType>&       types2
+   const QStringList&        entries1,
+   const QStringList&        entries2,
+   std::vector<DirDiffType>& types1,
+   std::vector<DirDiffType>& types2
 )
 {
    for ( uint ii = 0; ii < types1.size(); ++ii ) {
       if ( types1[ii] == UNKNOWN ) {
-         int index = searchEntry( entries2, entries1[ii] );
+         int index = entries2.findIndex( entries1[ii] );
          if ( index != -1 ) {
             types1[ ii ] = IDENTICAL;
             types2[ index ] = IDENTICAL;
@@ -338,78 +303,84 @@ void patchUpMissingTypes(
 //------------------------------------------------------------------------------
 //
 void buildSolelyFromOutput(
-   FILE*                     fp,
-   std::ostream&             errors,
-   const char*               path1,
+   FILE*                     fout,
+   QTextOStream&             errors,
+   const QString&            path1,
    XxBuffer*                 buffer1,
-   const char*               path2,
+   const QString&            path2,
    XxBuffer*                 buffer2,
    std::vector<DirDiffType>& types1,
    std::vector<DirDiffType>& types2
 )
 {
-   std::vector<const char*> entries1;
-   std::vector<const char*> entries2;
+   QStringList entries1;
+   QStringList entries2;
 
    types1.clear();
    types2.clear();
 
-   int len1 = ::strlen( path1 );
-   int len2 = ::strlen( path2 );
+   const int len1 = path1.length();
+   const int len2 = path2.length();
 
-   char buffer[BUFSIZ+1];
-   while ( fgets( buffer, BUFSIZ, fp ) != 0 ) {
+   QFile qfout;
+   qfout.open( IO_ReadOnly, fout );
+   QTextStream outputs( &qfout );
+
+   while ( true ) {
+      QString line = outputs.readLine();
+      if ( line.isNull() ) {
+         break;
+      }
 
       DirDiffType type;
-      std::string filename;
-      int onlyDir;
+      QString filename;
+      int onlyDir = -1;
       if ( parseDiffLine(
-         buffer, path1, len1, path2, len2, type, filename, onlyDir
+         line, path1, len1, path2, len2, type, filename, onlyDir
       ) == true ) {
          XX_LOCAL_TRACE( "ERROR" );
-         errors << "Diff error:" << std::endl;
-         errors << buffer << std::endl;
+         errors << "Diff error:" << endl;
+         errors << line << endl;
          continue;
       }
 
 #ifdef LOCAL_TRACE
-      std::cout << buffer;
-      std::cout << typeString[ type ] << "   " 
-                << filename << "   "
-                << onlyDir << std::endl;
+      XX_TRACE( line 
+                << typeString[ type ] << "   " 
+                << filename.latin1() << "   "
+                << onlyDir );
       if ( type == UNKNOWN ) {
-         throw new XxInternalError( XX_INTERROR_PARAMS );
+         throw XxInternalError( XX_EXC_PARAMS );
       }
 #endif
 
-      const char* dupname1 = ::strdup( filename.c_str() );
-      const char* dupname2 = ::strdup( filename.c_str() );
       switch ( type ) {
 
          case IDENTICAL: 
          case DIFFER: 
          case COMMON_SUBDIR: {
-            entries1.push_back( dupname1 );
-            entries2.push_back( dupname2 );
+            entries1.append( filename );
+            entries2.append( filename );
             types1.push_back( type );
             types2.push_back( type );
          } break;
 
          case ONLY_IN: {
             if ( onlyDir == 0 ) {
-               entries1.push_back( dupname1 );
+               entries1.append( filename );
                types1.push_back( type );
             }
             else {
-               entries2.push_back( dupname2 );
+               entries2.append( filename );
                types2.push_back( type );
             }
          } break;
          default: {
-            throw new XxInternalError( XX_INTERROR_PARAMS );
+            throw XxInternalError( XX_EXC_PARAMS );
          }
       }
    }
+   qfout.close();
 
    // Build the buffers.
    buffer1->setDirectoryEntries( entries1 );
@@ -419,49 +390,68 @@ void buildSolelyFromOutput(
 //------------------------------------------------------------------------------
 //
 void buildAgainstReadDirectory(
-   FILE*                     fp,
-   std::ostream&             errors,
-   const char*               path1,
+   FILE*                     fout,
+   QTextOStream&             errors,
+   const QString&            path1,
    const XxBuffer*           buffer1,
-   const char*               path2,
+   const QString&            path2,
    const XxBuffer*           buffer2,
    std::vector<DirDiffType>& types1,
    std::vector<DirDiffType>& types2
 )
 {
-   const std::vector<const char*>& entries1 = buffer1->getDirectoryEntries();
-   const std::vector<const char*>& entries2 = buffer2->getDirectoryEntries();
+   const QStringList& entries1 = buffer1->getDirectoryEntries();
+   const QStringList& entries2 = buffer2->getDirectoryEntries();
+#ifdef LOCAL_TRACE
+   {
+      for ( QStringList::ConstIterator iter = entries1.begin();
+            iter != entries1.end(); ++iter ) {
+         XX_TRACE( (*iter).latin1() );
+      }
+      for ( QStringList::ConstIterator iter = entries2.begin();
+            iter != entries2.end(); ++iter ) {
+         XX_TRACE( (*iter).latin1() );
+      }
+   }
+#endif
 
    types1.clear();
    types2.clear();
-   types1.insert( types1.begin(), entries1.size(), UNKNOWN );
-   types2.insert( types2.begin(), entries2.size(), UNKNOWN );
+   types1.insert( types1.begin(), entries1.count(), UNKNOWN );
+   types2.insert( types2.begin(), entries2.count(), UNKNOWN );
 
-   int len1 = ::strlen( path1 ); 
-   int len2 = ::strlen( path2 );
+   const int len1 = path1.length();
+   const int len2 = path2.length();
 
-   char buffer[BUFSIZ+1];
-   while ( fgets( buffer, BUFSIZ, fp ) != 0 ) {
+   QFile qfout;
+   qfout.open( IO_ReadOnly, fout );
+   QTextStream outputs( &qfout );
+
+   while ( true ) {
+      QString line = outputs.readLine();
+      if ( line.isNull() ) {
+         break;
+      }
 
       DirDiffType type;
-      std::string filename;
-      int onlyDir;
+      QString filename;
+      int onlyDir = -1;
       if ( parseDiffLine(
-         buffer, path1, len1, path2, len2, type, filename, onlyDir
+         line, path1, len1, path2, len2, type, filename, onlyDir
       ) == true ) {
          XX_LOCAL_TRACE( "ERROR" );
-         errors << "Diff error:" << std::endl;
-         errors << buffer << std::endl;
+         errors << "Diff error:" << endl;
+         errors << line << endl;
          continue;
       }
 
 #ifdef LOCAL_TRACE
-      std::cout << buffer;
-      std::cout << typeString[ type ] << "   " 
-                << filename << "   "
-                << onlyDir << std::endl;
+      XX_TRACE( line
+                << typeString[ type ] << "   " 
+                << filename.latin1() << "   "
+                << onlyDir );
       if ( type == UNKNOWN ) {
-         throw new XxInternalError( XX_INTERROR_PARAMS );
+         throw XxInternalError( XX_EXC_PARAMS );
       }
 #endif
 
@@ -470,23 +460,24 @@ void buildAgainstReadDirectory(
          case IDENTICAL: 
          case DIFFER: 
          case COMMON_SUBDIR: {
-            setType( entries1, types1, filename.c_str(), type );
-            setType( entries2, types2, filename.c_str(), type );
+            setType( entries1, types1, filename, type );
+            setType( entries2, types2, filename, type );
          } break;
 
          case ONLY_IN: {
             if ( onlyDir == 0 ) {
-               setType( entries1, types1, filename.c_str(), ONLY_IN );
+               setType( entries1, types1, filename, ONLY_IN );
             }
             else {
-               setType( entries2, types2, filename.c_str(), ONLY_IN );
+               setType( entries2, types2, filename, ONLY_IN );
             }
          } break;
          default: {
-            throw new XxInternalError( XX_INTERROR_PARAMS );
+            throw XxInternalError( XX_EXC_PARAMS );
          }
       }
    }
+   qfout.close();
 
    {
       std::vector<DirDiffType>::const_iterator it1 = 
@@ -494,9 +485,8 @@ void buildAgainstReadDirectory(
       std::vector<DirDiffType>::const_iterator it2 = 
          std::find( types2.begin(), types2.end(), UNKNOWN );
       if ( it1 != types1.end() || it2 != types2.end() ) {
-         errors << "Forgotten files in directory diffs." << std::endl
-                << "Check your subordinate directory diff program." << std::endl
-                << std::ends;
+         errors << "Forgotten files in directory diffs." << endl
+                << "Check your subordinate directory diff program." << endl;
          
          // Patch it up, fallback somehow: for each file that is UNKNOWN in the
          // first array, if the file is available and UNKNOWN in the second,
@@ -510,7 +500,7 @@ void buildAgainstReadDirectory(
       }
    }
 #ifdef LOCAL_TRACE
-   std::cout << "Done patching." << std::endl;
+   XX_TRACE( "Done patching." );
 #endif
 }
 
@@ -548,53 +538,59 @@ XxBuilderDirs2::~XxBuilderDirs2()
 //------------------------------------------------------------------------------
 //
 std::auto_ptr<XxDiffs> XxBuilderDirs2::process( 
-   const char* command,
-   const char* path1,
-   XxBuffer*   buffer1,
-   const char* path2,
-   XxBuffer*   buffer2
+   const QString& command,
+   const QString& path1,
+   XxBuffer*      buffer1,
+   const QString& path2,
+   XxBuffer*      buffer2
 )
 {
-   // Spawn command.
-   const char* args[3];
-   args[0] = path1;
-   args[1] = path2;
-   args[2] = 0;
-   FILE* fp = XxUtil::spawnCommandWithOutput( command, args );
-   if ( fp == 0 ) {
-      throw new XxIoError;
+   initLines();
+
+   QStringList filenames;
+   filenames.append( path1 );
+   filenames.append( path2 );
+   const char** out_args;
+   XxUtil::splitArgs( command, filenames, out_args );
+
+   FILE* fout;
+   FILE* ferr;
+   XxUtil::spawnCommand( out_args, &fout, &ferr );
+   if ( fout == 0 || ferr == 0 ) {
+      throw XxIoError( XX_EXC_PARAMS );
    }
+   XxUtil::freeArgs( out_args );
 
    std::vector<DirDiffType> types1;
    std::vector<DirDiffType> types2;
 
-   std::ostringstream errors;
+   QTextOStream errors( &_errors );
 
    // Note: for now we don't support recursive diffs built against a directory.
    if ( _buildSolelyFromOutput || _isDiffRecursive ) {
       buildSolelyFromOutput(
-         fp, errors, path1, buffer1, path2, buffer2, types1, types2
+         fout, errors, path1, buffer1, path2, buffer2, types1, types2
       );
    }
    else {
       buildAgainstReadDirectory( 
-         fp, errors, path1, buffer1, path2, buffer2, types1, types2
+         fout, errors, path1, buffer1, path2, buffer2, types1, types2
       );
    }
+   ::fclose( fout );
 
 #ifdef LOCAL_TRACE
-   std::cout << "------------------------------" << std::endl;
-   for ( int ii = 0; ii < types1.size(); ++ii ) {
-      std::cout << typeString[ types1[ii] ] << std::endl;
+   XX_TRACE( "------------------------------" );
+   for ( unsigned int ii = 0; ii < types1.size(); ++ii ) {
+      XX_TRACE( typeString[ types1[ii] ] );
    }
-   std::cout << "------------------------------" << std::endl;
-   std::cout << "------------------------------" << std::endl;
-   for ( int ii = 0; ii < types2.size(); ++ii ) {
-      std::cout << typeString[ types2[ii] ] << std::endl;
+   XX_TRACE( "------------------------------" );
+   XX_TRACE( "------------------------------" );
+   for ( unsigned int ii = 0; ii < types2.size(); ++ii ) {
+      XX_TRACE( typeString[ types2[ii] ] );
    }
-   std::cout << "------------------------------" << std::endl;
+   XX_TRACE( "------------------------------" );
 #endif
-
       
    XxFln fline1 = 1;
    XxFln fline2 = 1;
@@ -608,7 +604,7 @@ std::auto_ptr<XxDiffs> XxBuilderDirs2::process(
       while ( it1 != types1.end() || it2 != types2.end() ) {
          if ( it1 == types1.end() ) {
             if ( *it2 != ONLY_IN ) {
-               throw new XxInternalError( XX_INTERROR_PARAMS );
+               throw XxInternalError( XX_EXC_PARAMS );
             }
             XxLine line( XxLine::INSERT_2, -1, fline2++ );
             addLine( line );
@@ -617,7 +613,7 @@ std::auto_ptr<XxDiffs> XxBuilderDirs2::process(
          }
          if ( it2 == types2.end() ) {
             if ( *it1 != ONLY_IN ) {
-               throw new XxInternalError( XX_INTERROR_PARAMS );
+               throw XxInternalError( XX_EXC_PARAMS );
             }
             XxLine line( XxLine::INSERT_1, fline1++, -1 );
             addLine( line );
@@ -637,7 +633,7 @@ std::auto_ptr<XxDiffs> XxBuilderDirs2::process(
          }
          else if ( *it1 == DIFFER ) {
             if ( *it2 != *it1 ) {
-               throw new XxInternalError( XX_INTERROR_PARAMS );
+               throw XxInternalError( XX_EXC_PARAMS );
             }
             XxLine::Type dtype = 
                _ignoreFileChanges == true ? XxLine::SAME : XxLine::DIFF_ALL;
@@ -648,7 +644,7 @@ std::auto_ptr<XxDiffs> XxBuilderDirs2::process(
          }
          else if ( *it1 == IDENTICAL ) {
             if ( *it2 != *it1 ) {
-               throw new XxInternalError( XX_INTERROR_PARAMS );
+               throw XxInternalError( XX_EXC_PARAMS );
             }
             XxLine line( XxLine::SAME, fline1++, fline2++ );
             addLine( line );
@@ -657,14 +653,13 @@ std::auto_ptr<XxDiffs> XxBuilderDirs2::process(
          }
          else if ( *it1 == COMMON_SUBDIR ) {
             if ( *it2 != *it1 ) {
-               throw new XxInternalError( XX_INTERROR_PARAMS );
+               throw XxInternalError( XX_EXC_PARAMS );
             }
             XxLine line( XxLine::DIRECTORIES, fline1++, fline2++ );
             addLine( line );
             ++it1;
             ++it2;
          }
-         // FIXME simplify the three cases above.
          else {
             XX_ASSERT( false );
          }
@@ -686,36 +681,49 @@ std::auto_ptr<XxDiffs> XxBuilderDirs2::process(
       }
    }
 
-   // Save error text.
-   _errors = errors.str();
-   XX_LOCAL_TRACE( "Errors" << _errors );
+   // Collect stderr.
+   QFile qferr;
+   qferr.open( IO_ReadOnly, ferr );
+   {
+      QTextStream errorss( &qferr );
+      QString errstr = errorss.read();
+      if ( !errstr.isNull() ) {
+         errors << errstr << endl;
+      }
+   }
+   qferr.close();
+   ::fclose( ferr );
+
+   // Saved error text.
+   errors << flush;
+   XX_LOCAL_TRACE( "Errors: " << _errors );
 
    // If we've read no lines and there are diff errors then blow off
    if ( ( fline1 == 1 ) && ( fline2 == 1 ) && hasErrors() ) {
+#ifndef WINDOWS
       int stat_loc;
       if ( wait( &stat_loc ) == -1 ) {
-         throw new XxIoError;
+         throw XxIoError( XX_EXC_PARAMS );
       }
       _status = (WIFEXITED(stat_loc)) ? (WEXITSTATUS(stat_loc)) : 2;
-      throw new XxIoError;
+      throw XxIoError( XX_EXC_PARAMS );
+#else
+      _status = 2;
+#endif
    }
 
+#ifndef WINDOWS
    int stat_loc;
    if ( wait( &stat_loc ) == -1 ) {
-      throw new XxIoError;
+      throw XxIoError( XX_EXC_PARAMS );
    }
    _status = (WIFEXITED(stat_loc)) ? (WEXITSTATUS(stat_loc)) : 2;
+#else
+   _status = 2;
+#endif
 
    std::auto_ptr<XxDiffs> ap( new XxDiffs( _lines, true ) );
    return ap;
 }
 
-//------------------------------------------------------------------------------
-//
-void XxBuilderDirs2::addLine( const XxLine& line )
-{
-   XX_LOCAL_TRACE( "AddLine: " << line );
-   _lines.push_back( line );
-}
-
 XX_NAMESPACE_END
