@@ -1,6 +1,6 @@
 /******************************************************************************\
- * $Id: builderFiles2.cpp,v 1.3 2001/05/22 07:30:19 blais Exp $
- * $Date: 2001/05/22 07:30:19 $
+ * $Id: builderFiles2.cpp,v 1.30 2002/02/23 17:43:56 blais Exp $
+ * $Date: 2002/02/23 17:43:56 $
  *
  * Copyright (C) 1999-2001  Martin Blais <blais@iro.umontreal.ca>
  *
@@ -18,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *****************************************************************************/
+ ******************************************************************************/
 
 /*==============================================================================
  * EXTERNAL DECLARATIONS
@@ -28,14 +28,24 @@
 #include <exceptions.h>
 #include <diffs.h>
 #include <util.h>
+#include <diffutils.h>
 
+#include <qstring.h>
+#include <qtextstream.h>
+#include <qfile.h>
+
+#include <stdexcept>
 #include <stdio.h>
 #include <stdlib.h>
-#include <sstream>
+
 #include <unistd.h>
+#ifndef WINDOWS
+#include <sys/wait.h>
+#else
+#endif
+
 #include <iostream>
 #include <sys/types.h>
-#include <sys/wait.h>
 
 //#define LOCAL_TRACE
 #ifdef LOCAL_TRACE
@@ -44,6 +54,8 @@
 #define XX_LOCAL_TRACE(x)
 #endif
 
+#include <memory>
+
 namespace {
 XX_NAMESPACE_USING
 
@@ -54,85 +66,91 @@ XX_NAMESPACE_USING
 //------------------------------------------------------------------------------
 //
 bool parseDiffLine( 
-   XxLine::Type& type,
-   const char*   buf, 
-   XxFln&        f1n1, 
-   XxFln&        f1n2, 
-   XxFln&        f2n1, 
-   XxFln&        f2n2 
+   XxLine::Type&  type,
+   const QString& line,
+   XxFln&         f1n1, 
+   XxFln&         f1n2, 
+   XxFln&         f2n1, 
+   XxFln&         f2n2 
 )
 {
    /* 
     * this code taken from "ediff.c" by David MacKenzie, a published,
     * uncopyrighted program to translate diff output into plain English 
     */ 
+   const char* buf = line.latin1();
 
    bool error = true;
    if ( ( buf[0] == '<' ) || ( buf[0] == '>' ) || ( buf[0] == '-' ) ) {
       type = XxLine::SAME;
       error = false;
    }
-   else if ( sscanf( buf, "%d,%dc%d,%d\n", &f1n1, &f1n2, &f2n1, &f2n2 ) == 4 ) {
+   else if ( sscanf( buf, "%d,%dc%d,%d", &f1n1, &f1n2, &f2n1, &f2n2 ) == 4 ) {
       type = XxLine::DIFF_ALL;
       error = false;
    }
-   else if ( sscanf( buf, "%d,%dc%d\n", &f1n1, &f1n2, &f2n1) == 3 ) {
+   else if ( sscanf( buf, "%d,%dc%d", &f1n1, &f1n2, &f2n1) == 3 ) {
       f2n2 = f2n1;
       type = XxLine::DIFF_ALL;
       error = false;
    }
-   else if ( sscanf( buf, "%dc%d,%d\n", &f1n1, &f2n1, &f2n2 ) == 3 ) {
+   else if ( sscanf( buf, "%dc%d,%d", &f1n1, &f2n1, &f2n2 ) == 3 ) {
       f1n2 = f1n1;
       type = XxLine::DIFF_ALL;
       error = false;
    }
-   else if ( sscanf( buf, "%dc%d\n", &f1n1, &f2n1 ) == 2 ) {
+   else if ( sscanf( buf, "%dc%d", &f1n1, &f2n1 ) == 2 ) {
       f2n2 = f2n1;
       f1n2 = f1n1;
       type = XxLine::DIFF_ALL;
       error = false;
    }
-   else if ( sscanf( buf, "%d,%dd%d\n", &f1n1, &f1n2, &f2n1 ) == 3 ) {
+   else if ( sscanf( buf, "%d,%dd%d", &f1n1, &f1n2, &f2n1 ) == 3 ) {
       f2n2 = f2n1;
       type = XxLine::INSERT_1;
       error = false;
    }
-   else if ( sscanf( buf, "%dd%d\n", &f1n1, &f2n1 ) == 2 ) {
+   else if ( sscanf( buf, "%dd%d", &f1n1, &f2n1 ) == 2 ) {
       f2n2 = f2n1;
       f1n2 = f1n1;
       type = XxLine::INSERT_1;
       error = false;
    }
-   else if ( sscanf( buf, "%da%d,%d\n", &f1n1, &f2n1, &f2n2 ) == 3 ) {
+   else if ( sscanf( buf, "%da%d,%d", &f1n1, &f2n1, &f2n2 ) == 3 ) {
       f1n2 = f1n1;
       type = XxLine::INSERT_2;
       error = false;
    }
-   else if ( sscanf( buf, "%da%d\n", &f1n1, &f2n1 ) == 2 ) {
+   else if ( sscanf( buf, "%da%d", &f1n1, &f2n1 ) == 2 ) {
       f1n2 = f1n1;
       f2n2 = f2n1;
       type = XxLine::INSERT_2;
       error = false;
    }
-   // else
+   else if ( strncmp( buf, "\\ No newline at end of file", 27 ) == 0 ) {
+      type = XxLine::DIRECTORIES; /* Ignore this line. */
+      error = false;
+   }
+
    return error;
 }
 
+#ifndef NO_PARSE_DIFF_ERROR // Not needed for now.
+
 /*==============================================================================
  * CLASS XxParseDiffError
  *============================================================================*/
 
-class XxParseDiffError : public std::exception {
+class XxParseDiffError : public XxError,
+                         public std::runtime_error {
 
 public:
 
    /*----- member functions -----*/
 
-   // Default constructor.
-   XxParseDiffError();
-
    // Constructor with state.
    XxParseDiffError(
+      XX_EXC_PARAMS_DECL(file,line),
       const XxFln fline1, 
       const XxFln fline2, 
       const XxFln f1n1, 
@@ -141,63 +159,28 @@ public:
       const XxFln f2n2 
    );
 
-   // Destructor.
-   virtual ~XxParseDiffError() XX_THROW_NOTHING;
-
-   // See base class.
-   virtual const char* what() const XX_THROW_NOTHING;
-
-private:
-
-   /*----- data members -----*/
-
-   std::string _msg;
-
 };
 
-/*==============================================================================
- * CLASS XxParseDiffError
- *============================================================================*/
-
-//------------------------------------------------------------------------------
-//
-XxParseDiffError::XxParseDiffError()
-{
-   std::ostringstream oss;
-   oss << "Error parsing diff output." << std::ends;
-   _msg = oss.str();
-}
-
 //------------------------------------------------------------------------------
 //
 XxParseDiffError::XxParseDiffError(
+   XX_EXC_PARAMS_DECL(file,line),
    const XxFln fline1, 
    const XxFln fline2, 
    const XxFln f1n1, 
    const XxFln f1n2, 
    const XxFln f2n1, 
    const XxFln f2n2 
-)
+) :
+   XxError( file, line ),
+   std::runtime_error( "Parse diff output error." )
 {
-   std::ostringstream oss;
+   QTextStream oss( &_msg, IO_WriteOnly | IO_Append );
    oss << "Error parsing diff output: file1:" 
        << fline1 << " (" << f1n1 << "," << f1n2 << ")  file2: " 
-       << fline2 << " (" << f2n1 << "," << f2n2 << ")" << std::endl;
-   _msg = oss.str();
+       << fline2 << " (" << f2n1 << "," << f2n2 << ")" << endl;
 }
-
-//------------------------------------------------------------------------------
-//
-XxParseDiffError::~XxParseDiffError() XX_THROW_NOTHING
-{}
-
-//------------------------------------------------------------------------------
-//
-const char* XxParseDiffError::what() const XX_THROW_NOTHING
-{
-   return _msg.c_str();
-}
-
+#endif
 
 }
 
@@ -214,8 +197,9 @@ XX_NAMESPACE_BEGIN
 
 //------------------------------------------------------------------------------
 //
-XxBuilderFiles2::XxBuilderFiles2() :
-   XxBuilder()
+XxBuilderFiles2::XxBuilderFiles2( bool useInternalDiff ) :
+   XxBuilder(),
+   _useInternalDiff( useInternalDiff )
 {}
 
 //------------------------------------------------------------------------------
@@ -223,45 +207,77 @@ XxBuilderFiles2::XxBuilderFiles2() :
 XxBuilderFiles2::~XxBuilderFiles2()
 {}
 
-
 //------------------------------------------------------------------------------
 //
 std::auto_ptr<XxDiffs> XxBuilderFiles2::process( 
-   const char* command,
-   const char* path1,
-   const uint  nbLines1,
-   const char* path2,
-   const uint  nbLines2
+   const QString& command,
+   const QString& path1,
+   const uint     nbLines1,
+   const QString& path2,
+   const uint     nbLines2
 )
 {
-   const char* args[3];
-   args[0] = path1;
-   args[1] = path2;
-   args[2] = 0;
-   FILE* fp = XxUtil::spawnCommandWithOutput( command, args );
-   if ( fp == 0 ) {
-      throw new XxIoError;
+   initLines();
+
+//#define XX_INTERNAL_DIFFS
+
+   QStringList filenames;
+   filenames.append( path1 );
+   filenames.append( path2 );
+   const char** out_args;
+#ifdef XX_INTERNAL_DIFFS
+   int argc = 
+#endif
+   XxUtil::splitArgs( command, filenames, out_args );
+
+#ifndef XX_INTERNAL_DIFFS
+   FILE* fout;
+   FILE* ferr;
+   XxUtil::spawnCommand( out_args, &fout, &ferr );
+   if ( fout == 0 || ferr == 0 ) {
+      throw XxIoError( XX_EXC_PARAMS );
    }
+#else
+   std::auto_ptr<XxDiffutils> diffutils( new XxDiffutils );
+   diffutils->diff( argc, const_cast<char**>(out_args) );
+#endif
+   XxUtil::freeArgs( out_args );
 
    _curHunk = 0;
    XxFln fline1 = 1;
    XxFln fline2 = 1;
 
-   std::ostringstream errors;
-   char buffer[BUFSIZ+1];
+   QFile qfout;
+   qfout.open( IO_ReadOnly, fout );
+   QTextStream outputs( &qfout );
+
+   QTextOStream errors( &_errors );
    XxFln f1n1, f1n2, f2n1, f2n2;
-   while ( fgets( buffer, BUFSIZ, fp ) != 0 ) {
+
+   while ( true ) {
+#ifndef XX_INTERNAL_DIFFS
+      QString line = outputs.readLine();
+      if ( line.isNull() ) {
+         break;
+      }
+#else
+      QString line = diffutils->readLine();
+      if ( line.isNull() ) {
+         break;
+      }
+#endif
+
       XxLine::Type type;
-      if ( parseDiffLine( type, buffer, f1n1, f1n2, f2n1, f2n2 ) == true ) {
+      if ( parseDiffLine( type, line, f1n1, f1n2, f2n1, f2n2 ) == true ) {
          XX_LOCAL_TRACE( "ERROR" );
-         errors << "Diff error:" << std::endl;
-         errors << buffer << std::endl;
+         errors << "Diff error:" << endl;
+         errors << line << endl;
          continue;
       }
 
       switch ( type ) {
          case XxLine::INSERT_1: {
-            XX_LOCAL_TRACE( XxLine::mapToString( type ) );
+            XX_LOCAL_TRACE( XxLine::mapToString( type ).latin1() );
             XX_LOCAL_TRACE( "Output: f1n1=" << f1n1 << "  f1n2=" << f1n2 <<
                             "  f2n1=" << f2n1 << "  f2n2=" << f2n2 );
 
@@ -282,7 +298,7 @@ std::auto_ptr<XxDiffs> XxBuilderFiles2::process(
          } break;
 
          case XxLine::INSERT_2: {
-            XX_LOCAL_TRACE( XxLine::mapToString( type ) );
+            XX_LOCAL_TRACE( XxLine::mapToString( type ).latin1() );
             XX_LOCAL_TRACE( "Output: f1n1=" << f1n1 << "  f1n2=" << f1n2 <<
                             "  f2n1=" << f2n1 << "  f2n2=" << f2n2 );
 
@@ -303,7 +319,7 @@ std::auto_ptr<XxDiffs> XxBuilderFiles2::process(
          } break;
 
          case XxLine::DIFF_ALL: {
-            XX_LOCAL_TRACE( XxLine::mapToString( type ) );
+            XX_LOCAL_TRACE( XxLine::mapToString( type ).latin1() );
             XX_LOCAL_TRACE( "Output: f1n1=" << f1n1 << "  f1n2=" << f1n2 <<
                             "  f2n1=" << f2n1 << "  f2n2=" << f2n2 );
 
@@ -312,7 +328,7 @@ std::auto_ptr<XxDiffs> XxBuilderFiles2::process(
             // line changes (-B)!  Disabled.
             //
 	    // if ( (f1n1 - fline1) != (f2n1 - fline2) ) {
-            //    throw new XxParseDiffError( 
+            //    throw XxParseDiffError( 
             //       fline1, fline2, f1n1, f1n2, f2n1, f2n2 
             //    );
             // }
@@ -336,7 +352,11 @@ std::auto_ptr<XxDiffs> XxBuilderFiles2::process(
          } break;
 
          case XxLine::SAME: {
-            XX_LOCAL_TRACE( XxLine::mapToString( type ) );
+            XX_LOCAL_TRACE( XxLine::mapToString( type ).latin1() );
+         } break;
+
+         /* Used to ignore a line */
+         case XxLine::DIRECTORIES: {
          } break;
 
          case XxLine::DIFF_1:
@@ -348,41 +368,72 @@ std::auto_ptr<XxDiffs> XxBuilderFiles2::process(
          case XxLine::INSERT_3:
          case XxLine::DIFFDEL_1:
          case XxLine::DIFFDEL_2:
-         case XxLine::DIFFDEL_3:
-         case XxLine::DIRECTORIES: {
+         case XxLine::DIFFDEL_3: {
          }
 
+         case XxLine::NB_TYPES: {
+            XX_ABORT();
+         }
       }
    }
+   qfout.close();
 
-   // Save error text.
-   _errors = errors.str();
-   XX_LOCAL_TRACE( "Errors" << _errors );
+   // Collect stderr.
+   QFile qferr;
+   qferr.open( IO_ReadOnly, ferr );
+   {
+      QTextStream errorss( &qferr );
+      QString errstr = errorss.read();
+      if ( !errstr.isNull() ) {
+         errors << errstr << endl;
+      }
+   }
+   qferr.close();
+
+   // Saved error text.
+   errors << flush;
+   XX_LOCAL_TRACE( "Errors: " << _errors );
 
    // If we've read no lines and there are diff errors then blow off
    if ( ( fline1 == 1 ) && ( fline2 == 1 ) && hasErrors() ) {
+#if !defined(XX_INTERNAL_DIFFS) && !defined(WINDOWS)
       int stat_loc;
       if ( wait( &stat_loc ) == -1 ) {
-         throw new XxIoError;
+         throw XxIoError( XX_EXC_PARAMS );
       }
       _status = (WIFEXITED(stat_loc)) ? (WEXITSTATUS(stat_loc)) : 2;
-      throw new XxIoError;
+#else
+      _status = 2;
+#endif
+      throw XxError( XX_EXC_PARAMS, _errors );
    }
 
    // Add final ignore region if present.
    uint nbRemainingLines = nbLines1 + 1 - fline1;
    if ( nbRemainingLines != nbLines2 + 1 - fline2 ) {
-      throw new XxIoError;
+      throw XxError( XX_EXC_PARAMS, _errors );
    }
    if ( nbRemainingLines > 0 ) { 
       createIgnoreBlock( fline1, fline2, nbRemainingLines );
    }
 
+#if !defined(XX_INTERNAL_DIFFS) && !defined(WINDOWS)
    int stat_loc;
    if ( wait( &stat_loc ) == -1 ) {
-      throw new XxIoError;
+      throw XxIoError( XX_EXC_PARAMS );
    }
    _status = (WIFEXITED(stat_loc)) ? (WEXITSTATUS(stat_loc)) : 2;
+#else
+   _status = 2;
+#endif
+
+#ifndef XX_INTERNAL_DIFFS
+   ::fclose( fout );
+   ::fclose( ferr );
+#else
+   std::auto_ptr<XxDiffutils> null( 0 );
+   diffutils = null;
+#endif
 
    std::auto_ptr<XxDiffs> ap( new XxDiffs( _lines ) );
    return ap;
@@ -471,14 +522,4 @@ void XxBuilderFiles2::createInsertRightBlock(
    _curHunk++;
 }
 
-
-//------------------------------------------------------------------------------
-//
-void XxBuilderFiles2::addLine( const XxLine& line )
-{
-   XX_LOCAL_TRACE( "AddLine: " << line );
-   _lines.push_back( line );
-}
-
-
 XX_NAMESPACE_END
