@@ -1,6 +1,6 @@
 /******************************************************************************\
- * $Id: builderFiles3.cpp,v 1.3 2001/05/22 07:30:19 blais Exp $
- * $Date: 2001/05/22 07:30:19 $
+ * $Id: builderFiles3.cpp,v 1.21 2002/02/23 17:43:56 blais Exp $
+ * $Date: 2002/02/23 17:43:56 $
  *
  * Copyright (C) 1999-2001  Martin Blais <blais@iro.umontreal.ca>
  *
@@ -18,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *****************************************************************************/
+ ******************************************************************************/
 
 /*==============================================================================
  * EXTERNAL DECLARATIONS
@@ -29,14 +29,22 @@
 #include <diffs.h>
 #include <util.h>
 
+#include <qstring.h>
+#include <qtextstream.h>
+#include <qfile.h>
+
+#include <stdexcept>
 #include <stdio.h>
 #include <stdlib.h>
-#include <sstream>
+
 #include <unistd.h>
+#ifndef WINDOWS
+#include <sys/wait.h>
+#else
+#endif
+
 #include <iostream>
 #include <sys/types.h>
-#include <sys/wait.h>
-#include <string.h>
 
 //#define LOCAL_TRACE
 #ifdef LOCAL_TRACE
@@ -52,18 +60,16 @@ XX_NAMESPACE_USING
  * CLASS XxParseDiffError
  *============================================================================*/
 
-class XxParseDiffError : public std::exception {
+class XxParseDiffError : public XxError, 
+                         public std::runtime_error {
 
 public:
 
    /*----- member functions -----*/
 
-   // Default constructor.
-   XxParseDiffError();
-
    // Constructor with state.
    XxParseDiffError(
-      const int   lineNo, 
+      XX_EXC_PARAMS_DECL(file,line),
       const XxFln f1n1, 
       const XxFln f1n2, 
       const XxFln f2n1, 
@@ -72,78 +78,45 @@ public:
       const XxFln f3n2 
    );
 
-   // Destructor.
-   virtual ~XxParseDiffError() XX_THROW_NOTHING;
-
-   // See base class.
-   virtual const char* what() const XX_THROW_NOTHING;
-
-private:
-
-   /*----- data members -----*/
-
-   std::string _msg;
-
 };
 
-/*==============================================================================
- * CLASS XxParseDiffError
- *============================================================================*/
-
-//------------------------------------------------------------------------------
-//
-XxParseDiffError::XxParseDiffError()
-{
-   std::ostringstream oss;
-   oss << "Error parsing diff output." << std::ends;
-   _msg = oss.str();
-}
-
 //------------------------------------------------------------------------------
 //
 XxParseDiffError::XxParseDiffError(
-   int   lineNo,
+   XX_EXC_PARAMS_DECL(file,line),
    XxFln f1n1, 
    XxFln f1n2, 
    XxFln f2n1, 
    XxFln f2n2,
    XxFln f3n1, 
    XxFln f3n2
-)
+) :
+   XxError( file, line ),
+   std::runtime_error( "Parse diff output error." )
 {
-   std::ostringstream oss;
-   oss << "Error parsing diff3 output at line " << lineNo << ": file1:" 
+   QTextStream oss( &_msg, IO_WriteOnly | IO_Append );
+   oss << "Error parsing diff3 output:"
        << " (" << f1n1 << "," << f1n2 << ")  file2: " 
        << " (" << f2n1 << "," << f2n2 << ")  file3: " 
-       << " (" << f3n1 << "," << f3n2 << ")" << std::endl;
-   _msg = oss.str();
-}
-
-//------------------------------------------------------------------------------
-//
-XxParseDiffError::~XxParseDiffError() XX_THROW_NOTHING
-{}
-
-//------------------------------------------------------------------------------
-//
-const char* XxParseDiffError::what() const XX_THROW_NOTHING
-{
-   return _msg.c_str();
+       << " (" << f3n1 << "," << f3n2 << ")" << endl;
 }
 
+/*==============================================================================
+ * LOCAL FUNCTIONS
+ *============================================================================*/
 
 //------------------------------------------------------------------------------
 //
 bool parseDiffLine( 
-   XxLine::Type& type,
-   const char*   buf, 
-   int&          sno, 
-   XxFln&        f1n1, 
-   XxFln&        f1n2, 
-   XxFln&        f2n1, 
-   XxFln&        f2n2, 
-   XxFln&        f3n1, 
-   XxFln&        f3n2 
+   XxLine::Type&  type,
+   const QString& line, 
+   int&           sno, 
+   XxFln&         f1n1, 
+   XxFln&         f1n2, 
+   XxFln&         f2n1, 
+   XxFln&         f2n2, 
+   XxFln&         f3n1, 
+   XxFln&         f3n2 
 )
 {
    //
@@ -228,11 +201,13 @@ bool parseDiffLine(
  
 #define PARSE_CHECK(xxxx) \
   if ( !(xxxx) ) { \
-      throw new XxParseDiffError( __LINE__, \
+      throw XxParseDiffError( XX_EXC_PARAMS, \
          f1n1, f1n2, f2n1, f2n2, f3n1, f3n2 \
       ); \
   }
 
+   const char* buf = line.latin1();
+
    XX_LOCAL_TRACE( "" );
    XX_LOCAL_TRACE( "" );
    XX_LOCAL_TRACE( "" );
@@ -247,18 +222,18 @@ bool parseDiffLine(
       XX_LOCAL_TRACE( "recognized text line" );
       error = false;
    }
-   else if ( sscanf( buf, "====%d\n", &sno ) == 1 ) {
+   else if ( sscanf( buf, "====%d", &sno ) == 1 ) {
       XX_LOCAL_TRACE( "recognized numbered block start" );
       f1n1 = f1n2 = f2n1 = f2n2 = f3n1 = f3n2 = -1;
       error = false;
    }
-   else if ( strncmp( buf, "====\n", 5 ) == 0 ) {
+   else if ( strncmp( buf, "====", 4 ) == 0 ) {
       XX_LOCAL_TRACE( "recognized un-unumbered block start" );
       sno = -1;
       f1n1 = f1n2 = f2n1 = f2n2 = f3n1 = f3n2 = -1;
       error = false;
    }
-   else if ( sscanf( buf, "%d:%d,%d%[c]\n", &r, &n1, &n2, cbuf ) == 4 ) {
+   else if ( sscanf( buf, "%d:%d,%d%[c]", &r, &n1, &n2, cbuf ) == 4 ) {
       XX_LOCAL_TRACE( "recognized range change" );
       switch ( r ) {
          case 1: {
@@ -277,7 +252,7 @@ bool parseDiffLine(
       checkForCompletion = true;
       error = false;
    }
-   else if ( sscanf( buf, "%d:%d%[c]\n", &r, &n1, cbuf ) == 3 ) {
+   else if ( sscanf( buf, "%d:%d%[c]", &r, &n1, cbuf ) == 3 ) {
       XX_LOCAL_TRACE( "recognized one-line change" );
       switch ( r ) {
          case 1: {
@@ -296,7 +271,7 @@ bool parseDiffLine(
       checkForCompletion = true;
       error = false;
    }
-   else if ( sscanf( buf, "%d:%d%[a]\n", &r, &n1, cbuf ) == 3 ) {
+   else if ( sscanf( buf, "%d:%d%[a]", &r, &n1, cbuf ) == 3 ) {
       XX_LOCAL_TRACE( "recognized append" );
       switch ( r ) {
          case 1: {
@@ -315,6 +290,10 @@ bool parseDiffLine(
       checkForCompletion = true;
       error = false;
    }
+   else if ( strncmp( buf, "\\ No newline at end of file", 27 ) == 0 ) {
+      XX_LOCAL_TRACE( "recognized incomplete line marker" );
+      error = false;
+   }
    else {
       PARSE_CHECK( false );
    }
@@ -427,24 +406,31 @@ XxBuilderFiles3::~XxBuilderFiles3()
 //------------------------------------------------------------------------------
 //
 std::auto_ptr<XxDiffs> XxBuilderFiles3::process( 
-   const char* command,
-   const char* path1,
-   const uint  nbLines1,
-   const char* path2,
-   const uint  nbLines2,
-   const char* path3,
-   const uint  nbLines3
+   const QString& command,
+   const QString& path1,
+   const uint     nbLines1,
+   const QString& path2,
+   const uint     nbLines2,
+   const QString& path3,
+   const uint     nbLines3
 )
 {
-   const char* args[4];
-   args[0] = path1;
-   args[1] = path2;
-   args[2] = path3;
-   args[3] = 0;
-   FILE* fp = XxUtil::spawnCommandWithOutput( command, args );
-   if ( fp == 0 ) {
-      throw new XxIoError;
+   initLines();
+
+   QStringList filenames;
+   filenames.append( path1 );
+   filenames.append( path2 );
+   filenames.append( path3 );
+   const char** out_args;
+   XxUtil::splitArgs( command, filenames, out_args );
+
+   FILE* fout;
+   FILE* ferr;
+   XxUtil::spawnCommand( out_args, &fout, &ferr );
+   if ( fout == 0 || ferr == 0 ) {
+      throw XxIoError( XX_EXC_PARAMS );
    }
+   XxUtil::freeArgs( out_args );
 
    _curHunk = 0;
    XxFln fline1 = 1;
@@ -452,29 +438,38 @@ std::auto_ptr<XxDiffs> XxBuilderFiles3::process(
    XxFln fline3 = 1;
 
    bool foundDifferences = false;
-   std::ostringstream errors;
-   char buffer[BUFSIZ+1];
+   QTextOStream errors( &_errors );
    int sno;
    XxFln f1n1, f1n2, f2n1, f2n2, f3n1, f3n2;
-   while ( fgets( buffer, BUFSIZ, fp ) != 0 ) {
+
+   QFile qfout;
+   qfout.open( IO_ReadOnly, fout );
+   QTextStream outputs( &qfout );
+
+   while ( true ) {
+      QString line = outputs.readLine();
+      if ( line.isNull() ) {
+         break;
+      }
+
       XxLine::Type type;
-      if ( parseDiffLine( type, buffer,
+      if ( parseDiffLine( type, line,
                           sno, f1n1, f1n2, f2n1, f2n2, f3n1, f3n2 ) == true ) {
          XX_LOCAL_TRACE( "ERROR" );
-         errors << "Diff error:" << std::endl;
-         errors << buffer << std::endl;
+         errors << "Diff error:" << endl;
+         errors << line << endl;
          continue;
       }
 
 #ifdef XX_DEBUG
       XX_LOCAL_TRACE( "ParseDiffLine results: " );
-      XX_LOCAL_TRACE( XxLine::mapToString( type ) );
+      XX_LOCAL_TRACE( XxLine::mapToString( type ).latin1() );
       XX_LOCAL_TRACE( "  sno=" << sno );
       XX_LOCAL_TRACE( "  f1n1=" << f1n1 << "  f1n2=" << f1n2 );
       XX_LOCAL_TRACE( "  f2n1=" << f2n1 << "  f2n2=" << f2n2 );
       XX_LOCAL_TRACE( "  f3n1=" << f3n1 << "  f3n2=" << f3n2 );
 #endif
-
+      
       if ( type != XxLine::SAME ) {
          foundDifferences = true;
          int fsize1 = f1n1 - fline1;
@@ -530,43 +525,66 @@ std::auto_ptr<XxDiffs> XxBuilderFiles3::process(
          _curHunk++;
       }
    }
+   qfout.close();
    
-   // Save error text.
-   _errors = errors.str();
-   XX_LOCAL_TRACE( "Errors" << _errors );
+   // Collect stderr.
+   QFile qferr;
+   qferr.open( IO_ReadOnly, ferr );
+   {
+      QTextStream errorss( &qferr );
+      QString errstr = errorss.read();
+      if ( !errstr.isNull() ) {
+         errors << errstr << endl;
+      }
+   }
+   qferr.close();
+
+   // Saved error text.
+   errors << flush;
+   XX_LOCAL_TRACE( "Errors: " << _errors );
 
    // If we've read no lines and there are diff errors then blow off
    if ( ( fline1 == 1 ) && ( fline2 == 1 ) && ( fline3 == 1 ) &&
         hasErrors() ) {
+#ifndef WINDOWS
       int stat_loc;
       if ( wait( &stat_loc ) == -1 ) {
-         throw new XxIoError;
+         throw XxIoError( XX_EXC_PARAMS );
       }
       _status = (WIFEXITED(stat_loc)) ? (WEXITSTATUS(stat_loc)) : 2;
-      throw new XxIoError;
+      throw XxError( XX_EXC_PARAMS, _errors );
+#else
+      _status = 2;
+#endif
    }
 
    // Add final ignore region if present.
    uint nbRemainingLines = nbLines1 + 1 - fline1;
    if ( nbRemainingLines != nbLines2 + 1 - fline2 ||
         nbRemainingLines != nbLines3 + 1 - fline3 ) {
-      throw new XxIoError;
+      throw XxError( XX_EXC_PARAMS, _errors );
    }
    if ( nbRemainingLines > 0 ) { 
       createIgnoreBlock( fline1, fline2, fline3, int(nbRemainingLines) );
    }
 
+#ifndef WINDOWS
    int stat_loc;
    if ( wait( &stat_loc ) == -1 ) {
-      throw new XxIoError;
+      throw XxIoError( XX_EXC_PARAMS );
    }
    _status = (WIFEXITED(stat_loc)) ? (WEXITSTATUS(stat_loc)) : 2;
-
+#else
+   _status = 2;
+#endif
    // Fix for deficient non-GNU diff3.
    if ( _status == 0 && foundDifferences == true ) {
       _status = 1;
    }
 
+   ::fclose( fout );
+   ::fclose( ferr );
+
    std::auto_ptr<XxDiffs> ap( new XxDiffs( _lines ) );
    return ap;
 }
@@ -588,13 +606,4 @@ void XxBuilderFiles3::createIgnoreBlock(
    _curHunk++;
 }
 
-//------------------------------------------------------------------------------
-//
-void XxBuilderFiles3::addLine( const XxLine& line )
-{
-   XX_LOCAL_TRACE( "AddLine: " << line );
-   _lines.push_back( line );
-}
-
-
 XX_NAMESPACE_END
