@@ -1,6 +1,6 @@
 /******************************************************************************\
- * $Id: diffs.h,v 1.10 2001/06/08 16:04:58 blais Exp $
- * $Date: 2001/06/08 16:04:58 $
+ * $Id: diffs.h,v 1.27 2002/02/23 17:43:56 blais Exp $
+ * $Date: 2002/02/23 17:43:56 $
  *
  * Copyright (C) 1999-2001  Martin Blais <blais@iro.umontreal.ca>
  *
@@ -18,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *****************************************************************************/
+ ******************************************************************************/
 
 #ifndef INCL_XXDIFF_DIFFS
 #define INCL_XXDIFF_DIFFS
@@ -35,6 +35,10 @@
 #include <defs.h>
 #endif
 
+#ifndef INCL_XXDIFF_RESOURCES
+#include <resources.h>
+#endif
+
 #ifndef INCL_QT_QOBJECT
 #include <qobject.h>
 #define INCL_QT_QOBJECT
@@ -55,18 +59,14 @@
 #define INCL_STD_IOSFWD
 #endif
 
-
 XX_NAMESPACE_BEGIN
 
-
 /*==============================================================================
  * FORWARD DECLARATIONS
  *============================================================================*/
 
 class XxBuffer;
 
-
-
 /*==============================================================================
  * CLASS XxDiffs
  *============================================================================*/
@@ -106,18 +106,33 @@ public:
    XxDiffs();
 
    // Constructor.
-   XxDiffs( std::vector<XxLine>& lines, bool isDirectoryDiff = false );
+   // Validate checks that the line numbers are sequential.
+   //
+   // Important note: the given lines array must have a dummy line inserted at
+   // index 0. In other words, display line 1 should be at index 1. This is
+   // simply to avoid subtracting 1 in a bunch of places and to follow the
+   // external convention for display line numbers.
+   XxDiffs(
+      std::vector<XxLine>& lines,
+      bool                 isDirectoryDiff = false,
+      bool                 validate = true
+   );
 
    // Destructor.
    virtual ~XxDiffs();
 
    // Returns the number of display lines.
-   uint getNbLines() const;
+   XxDln getNbLines() const;
 
    // Returns the number of lines that have selected content (i.e. either just
    // have content, or have a selected side with content).  Counts the number of
    // times we encountered new unselected changes (for merged view).
-   uint getNbLinesWithContent( uint& unselectedChanges ) const;
+   XxDln getNbLinesWithContent( uint& unselectedChanges ) const;
+
+   // For a specific file, returns the number of lines with text for that
+   // file. We cannot trust the number of lines that the buffer returns, since
+   // not all lines from that buffer may be displayed.
+   XxFln getNbLinesWithText( const XxFno fno ) const;
 
    // Moves backwards nbLines lines from a specified starting line of the diffs,
    // counting only visible lines according to the current selection, counting
@@ -202,7 +217,7 @@ public:
    // Returns the closest (floor) file line for the given display line.  The
    // actuallyEmpty parameter is set to true if that line is actually empty for
    // that file no.
-   XxFln getFileLine( 
+   XxFln getBufferLine( 
       XxFno no,
       XxDln lineNo,
       bool& actuallyEmpty
@@ -220,20 +235,24 @@ public:
    // Returns false if there are still some unselected lines.
    bool isAllSelected() const;
 
+   // Counts the remaining unselected hunks.
+   uint countRemainingUnselected() const;
+
    // Save merged with selected regions.  If there are still some unselected
    // regions it will return false but save anyway.
    bool save( 
-      std::ostream&                  os, 
+      const XxResources&             resources,
+      QTextStream&                   os, 
       const std::auto_ptr<XxBuffer>* files,
       const bool                     useConditionals,
       const bool                     removeEmptyConditionals,
-      const std::string              conditionals[3]
+      const QString                  conditionals[3]
    ) const;
 
    // Save selected lines with an appropriate prefix.  Returns true if there
    // were some regions saved (false thus means an empty file).
    bool saveSelectedOnly( 
-      std::ostream& os, 
+      QTextStream&                   os, 
       const std::auto_ptr<XxBuffer>* files
    ) const;
 
@@ -241,7 +260,7 @@ public:
    // diff lines where the text appears in either, and for each diff line, in
    // which file the line is found.
    void search( 
-      const char*                    searchText,
+      const QString&                 searchText,
       const int                      nbFiles,
       const std::auto_ptr<XxBuffer>* files
    );
@@ -257,13 +276,16 @@ public:
 
    // Initialize horizontal diffs for lines.
    void initializeHorizontalDiffs(
+      const XxResources&             resources,
       const std::auto_ptr<XxBuffer>* files,
-      const bool                   force = false
+      const bool                     force = false
    );
 
    // Automatically merge by performing selections on file1, assuming order
    // "mine older yours", and leaving conflicts unselected.
-   void merge();
+   //
+   // Return result as diff3 merge feature would.
+   int merge( uint nbFiles );
 
    // Check if all selections are of the specified no.  If so return true.
    bool checkSelections( const XxLine::Selection sel ) const;
@@ -271,6 +293,15 @@ public:
    // Returns true if this diffs is a directory diffs.
    bool isDirectoryDiff() const;
 
+   // Re-index the buffers so that the lines they believe they have seem to be
+   // contiguous, and adjust the diffs accordingly. This method modifies both
+   // the buffers as well as this instance. This is used for unmerge.
+   void reindex( 
+      const std::auto_ptr<XxBuffer>& file1,
+      const std::auto_ptr<XxBuffer>& file2
+   );
+
+
    // Dump debug output.
    std::ostream& dump( std::ostream& ) const;
 
@@ -302,7 +333,7 @@ private:
    ) const;
 
    // Make really sure the line numbers are sequential.  This is zealous and
-   // that's how I like it.
+   // that's how I like it. Self-defensive programming.
    void validateLineNumbers() const;
 
    /*----- data members -----*/
@@ -312,7 +343,8 @@ private:
 
    std::vector< SearchResult > _searchResults;
 
-   bool                        _initializedHorizontalDiffs;
+   XxHordiff                   _lastInitializedDiffType;
+   uint                        _lastInitializedContext;
    bool                        _isDirectoryDiff;
    bool                        _dirty;
 
