@@ -1,6 +1,6 @@
 /******************************************************************************\
- * $Id: searchDialog.cpp,v 1.5 2001/05/22 07:30:19 blais Exp $
- * $Date: 2001/05/22 07:30:19 $
+ * $Id: searchDialog.cpp,v 1.11 2001/11/30 07:21:57 blais Exp $
+ * $Date: 2001/11/30 07:21:57 $
  *
  * Copyright (C) 1999-2001  Martin Blais <blais@iro.umontreal.ca>
  *
@@ -18,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- *****************************************************************************/
+ ******************************************************************************/
 
 /*==============================================================================
  * EXTERNAL DECLARATIONS
@@ -70,9 +70,9 @@ XxSearchDialog::XxSearchDialog(
 void XxSearchDialog::onApply()
 {
    XxDiffs* diffs = _app->getDiffs();
-   const char* searchText = _lineeditSearchString->text().ascii();
-   if ( diffs != 0 && searchText != 0 ) {
-      diffs->search( searchText, _app->getNbFiles(), _app->getFiles() );
+   const QString searchText = _lineeditSearchString->text();
+   if ( diffs != 0 && !searchText.isEmpty() ) {
+      diffs->search( searchText, _app->getNbFiles(), _app->getBuffers() );
    }
 }
 
@@ -80,8 +80,8 @@ void XxSearchDialog::onApply()
 //
 void XxSearchDialog::onGotoLine()
 {
-   const char* glinetext = _lineeditGotoLine->text().ascii();
-   int gline = atoi( glinetext );
+   bool ok;
+   int gline = _lineeditGotoLine->text().toInt( &ok );
    
    static int ind[3] = { 0, 2, 1 };
    XX_ASSERT( _comboGotoWhichFile->currentItem() < 3 );
@@ -95,4 +95,13 @@ void XxSearchDialog::onGotoLine()
    _app->setCursorLine( dline );
 }
 
+//------------------------------------------------------------------------------
+//
+void XxSearchDialog::show()
+{
+   _lineeditSearchString->setFocus();
+   BaseClass::show();
+}
+
 XX_NAMESPACE_END
+
